module kriging_mod

  use exp_cov_mod
  use invert_matrix_mod
  use mat_mult_mod

  implicit none

  real,private        :: nugget,range,sill,alpha,power ! parameters for the variogram need to be imported
  integer,private     :: niter

  contains

  subroutine ordinary_kriging_init(nugget_in,range_in,sill_in,alpha_in,power_in,niter_in)
    real               :: nugget_in,range_in,sill_in,alpha_in,power_in
    integer            :: niter_in

    nugget = nugget_in
    range = range_in
    sill = sill_in
    alpha = alpha_in
    niter = niter_in
    power = power_in
  end subroutine

  function ordinary_kriging(samples_position,samples_values,new_position,error_stand_dev) result(kriged_value)
    real,dimension(:),intent(in)       :: samples_values
    integer,dimension(:,:),intent(in)  :: samples_position
    real,dimension(:,:),allocatable    :: distance_matrix,cov_matrix
    integer,dimension(2),intent(in)    :: new_position
    real,dimension(:),allocatable      :: weight_vector,distance_vector,cov_vector,test_vector,diff
    integer                            :: nb_samples,i,j,k,flag
    real                               :: kriged_value,resd,norm_cov,ratio,lagrange,temp_sum,error_stand_dev
    integer                            :: stats

    ! variables allocation
    nb_samples = size(samples_position,1)
    allocate(distance_matrix(nb_samples,nb_samples))
    allocate(distance_vector(nb_samples))
    allocate(cov_matrix(nb_samples+1,nb_samples+1))
    allocate(weight_vector(nb_samples+1))
    allocate(cov_vector(nb_samples+1))
    allocate(test_vector(nb_samples+1))
    allocate(diff(nb_samples+1))

    ! calculating the distance matrix 
    distance_matrix=0.
    do i=1,nb_samples-1
      do j=i+1,nb_samples
        do k=1,2 ! for 2D 
          distance_matrix(i,j) = distance_matrix(i,j) + (1.0 * samples_position(i,k)-1.0 * samples_position(j,k))**2
        end do
        distance_matrix(i,j) = sqrt(distance_matrix(i,j))
        distance_matrix(j,i) = distance_matrix(i,j)
      end do
    end do

    ! calculating the covariance matrix
    cov_matrix = 0.
    do i=1,nb_samples
      do j=i,nb_samples
        cov_matrix(i,j) = get_covariance(distance_matrix(i,j),nugget,range,sill,alpha,power)
        cov_matrix(j,i) = cov_matrix(i,j)
      end do
    end do
   
    !write(0,*), "nb", nb_samples

    do i=1,nb_samples
      cov_matrix(i,nb_samples+1)=1.
      cov_matrix(nb_samples+1,i)=1.
    end do

    !do i=1,nb_samples+1
    !  do j=1,nb_samples+1
    !    write(0,*) "cov_matrix", cov_matrix(i,j)
    !  end do
    !end do

    ! calculating distance vector
    distance_vector = 0.
    do i=1,nb_samples
      do j=1,2
      distance_vector(i) = distance_vector(i) + (1.0 * new_position(j)-1.0 * samples_position(i,j))**2
      end do
      distance_vector(i) = sqrt(distance_vector(i))
    end do
   
    ! calculating the covariance vector (covariance between krigged point and sample points)
    cov_vector = 0.
    cov_vector(nb_samples+1)=1.
    do i=1,nb_samples
      cov_vector(i) = get_covariance(distance_vector(i),nugget,range,sill,alpha,power)
    end do

    !!!!!! calculating norm of covariance vector
    !norm_cov = 0.
    !do i=1,size(cov_vector)
    !  norm_cov = norm_cov + cov_vector(i)*cov_vector(i)
    !end do
    !norm_cov = sqrt(norm_cov)

    !do i=1,size(cov_vector)
    !  write(0,*), "cov vector", cov_vector(i)
    !end do

    ! inverting the covariance matrix
    call invert_matrix(cov_matrix,weight_vector,cov_vector,niter)

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!! checking the inversion by comparing the reconstructed data vs. real data !!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    !call mat_mult_init(nb_samples+1,cov_matrix)
    !stats = mat_mult_op(.false.,.false.,weight_vector,test_vector)
    !do i=1,size(weight_vector)
    !  write(0,*), "rec vector",i,test_vector(i)
    !  write(0,*), "cov vector",i,cov_vector(i)
    !end do

    !write(0,*) "nb samples",nb_samples

    !do i=1,size(weight_vector)
    !  write(0,*) "matrix(n+1,",i,")",cov_matrix(nb_samples+1,i)
    !end do

    ! calculating the last residuals
    !diff = 0.
    !resd = 0.
    !do i=1,nb_samples+1
    !  diff(i) = test_vector(i)-cov_vector(i)
    !end do
    !do i=1,nb_samples+1
    !  resd = resd + diff(i)*diff(i)
    !end do
    !resd = sqrt(resd)
    !ratio = resd/norm_cov

    !write(0,*) "residuals",resd
    !write(0,*) "norm cov", norm_cov
    !write(0,*) "ratio", ratio
    !call dealloc_matrix()

    !do i=1,size(weight_vector)
    !  write(0,*) "weight vector",i, weight_vector(i)
    !end do
    !write(0,*) "sum of weights", sum(weight_vector) - weight_vector(size(weight_vector))

    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    ! calculate kriged value
    kriged_value = 0.
    temp_sum = 0.

    do i=1,nb_samples
      kriged_value = kriged_value + weight_vector(i) * samples_values(i)
      temp_sum = temp_sum + weight_vector(i) * cov_vector(i) 
    end do

    lagrange = weight_vector(nb_samples+1)
    error_stand_dev = sill - lagrange - temp_sum
    error_stand_dev = sqrt(error_stand_dev)
    !write(0,*) "kriged value", kriged_value

  end function


end module
