module stats_mod

  implicit none

  contains

  ! building gaussian cdf - takes a 2D dataset as input and spits out a Gaussian cdf
  function gaussian_cdf(o_cdf,d_cdf,n_cdf,mean,sigma) result(cdf)

    real,dimension(:),allocatable      :: cdf
    real                               :: o_samp,d_samp,x_min,x_max,o_cdf,d_cdf
    real                               :: x
    real                               :: temp_sum,new_term,sigma,mean
    integer                            :: i,j
    integer                            :: n1,n2,n_cdf,n
    real                               :: p_min,p_max

    allocate(cdf(n_cdf))    
    cdf = 0.
    x_min = o_cdf
    x_max = o_cdf + d_cdf * (n_cdf-1) 
    !write(0,*) "x min",x_min
    !write(0,*) "x max",x_max

    ! calculating the cdf values for Gaussian model
    do i = 1,n_cdf
      x = o_cdf + (i-1) * d_cdf
      x = (x-mean)/sigma ! change of variable x_tilde = (x-mean)/sigma 
      temp_sum = x
      new_term = x
      do j=1,200
        new_term = new_term * x * x / (2*j+1)
        temp_sum = temp_sum + new_term
      end do
      cdf(i) = 0.5 + temp_sum / sqrt(2*3.14) * exp(-x*x/2)
      !write(0,*) "cdf",i,cdf(i)
    end do 

  end function

  ! mc_gaussian generates a pseudo-random value "x_random" using a Gaussian CDF
  function mc_gaussian(cdf,o_cdf,d_cdf,n_cdf,vect,index_vect) result(x_random)
    real,dimension(:)           :: cdf
    real                        :: o_cdf,d_cdf,rand,p_min,p_max,x_random
    integer                     :: n_cdf,i
    integer,optional            :: index_vect
    real,dimension(:),optional  :: vect

    p_min = cdf(1)
    p_max = cdf(n_cdf)

    !write(0,*) "p min",p_min
    !write(0,*) "p min",p_max

    ! rand is going to be a number between 0 and 1. We set this number to an arbitrary low value and then we generate a random value between 0 and 1
    ! we keep on generating random values until the value of "rand" falls between p_min and p_max -> this is made to avoid outliers
    rand = p_min - 50.0 

    do while ((rand .lt. p_min ) .or. (rand .gt. p_max))
      !if (present(counter)) counter = counter + 1
      if ( present(vect) .and. present(index_vect)) then
        index_vect = index_vect + 1
        rand = vect(index_vect)
      else
        call random_number(rand)  
      end if    
    end do
    i = 1
    do while (cdf(i) .lt. rand)
      i = i + 1
    end do
    x_random = o_cdf + (i-1) * d_cdf
    
  end function

  function average(samples) result(avg)
    real,dimension(:,:)    :: samples
    integer                :: n1,n2,n,i,j
    real                   :: avg
    n1 = size(samples,1)
    n2 = size(samples,2)   
    n  = n1 * n2
    avg = 0.
    do i=1,n1
      do j=1,n2
        avg = avg + samples(i,j)
      end do
    end do
    avg = avg/(1.0 * n)    
  end function

  function stand_dev(samples,mean) result(sigma)
    real                   :: sigma,mean
    real,dimension(:,:)    :: samples
    integer                :: n1,n2,n,i,j

    n1 = size(samples,1)
    n2 = size(samples,2)   
    n  = n1 * n2
    sigma = 0.
   
    do i=1,n1
      do j=1,n2
        sigma = sigma + (samples(i,j)-mean) * (samples(i,j)-mean)
      end do
    end do
    sigma = sigma / (1.0 * n - 1.0)
    sigma = sqrt(sigma)
    
  end function

  function raw_cdf(semblance,origin,increment,vel_min,vel_max) result(cdf) ! gets the cdf for a 1D array
    real,dimension(:)                     :: semblance ! array can be the semblance values for instance
    real,dimension(:),allocatable         :: cdf
    integer                               :: i,n,total_nb_samples
    integer,dimension(:),allocatable      :: occurrence ! array containing the nb of occurrence for each velocity value evaluated by the semblance plot
    real                                  :: origin,increment,vel_min,vel_max,velocity,total_weight
    integer                               :: index_min,index_max

    n=size(semblance)
    allocate(occurrence(n))

    ! finding the index of the lowest allowed velocity ! it has to be greater than the lowest velocity scanned
    index_min=1
    velocity = origin 
    do while (velocity .lt. vel_min)
      index_min = index_min+1
      velocity = origin + (index_min-1)*increment
    end do

    ! finding the index of the highest allowed velocity !it has to be smaller than the highest velocity scanned
    index_max=n
    velocity = origin + (n-1)*increment
    do while (velocity .gt. vel_max)
      index_max = index_max-1
      velocity = velocity-increment
    end do

    ! finding the total weighting (ie sum of all semblance values for a given tau)
    total_weight=0.
    do i=index_min,index_max
      total_weight = total_weight + semblance(i)
    end do

    ! creating the cdf
    allocate(cdf(n))
    cdf=0.
    cdf(index_min)=semblance(index_min)
    do i=index_min+1,index_max
      cdf(i)=cdf(i-1) + semblance(i)
    end do
    cdf = cdf/total_weight
    do i=index_max,n
      cdf(i)=1.
    end do 
    
  end function


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! calculating mean for each tau (parameter estimation)
  function get_mean(array,origin,increment) result(mean)
    real,dimension(:)      :: array
    real                   :: mean
    integer                :: p,i
    real                   :: origin,increment,value

    p=size(array)
    mean=0.
    do i=1,p
      value=origin+(i-1)*increment
      mean=mean+value*array(i)
    end do
  end function

  ! calculating standard deviation for each tau (parameter estimation)
  function standard_dev(array,origin,increment) result(sigma)
    real,dimension(:)      :: array
    real                   :: sigma,mean
    integer                :: p,i
    integer                :: nb
    real                   :: origin,increment,value

    p=size(array)
    mean=get_mean(array,origin,increment)
    sigma=0.  
      do i=1,p
        value=origin+(i-1)*increment
        sigma=sigma+(value-mean)*(value-mean)*array(i)
      end do
      sigma=sqrt(sigma)
  end function

  ! building gaussian cdf for each tau for a given CMP gather 
  function build_gaussian_cdf(array,origin,increment,vel_min,vel_max) result(cdf)
    real,dimension(:),allocatable      :: cdf,velocity_array
    real,dimension(:)                  :: array ! semblance array 
    real                               :: origin,increment,vel_min,vel_max,mean,sigma
    real                               :: temp_sum,velocity,new_term,total_weight
    integer                            :: n,i,k,nval,index_min,index_max

    n=size(array)
    allocate(cdf(n))   
    cdf=0.

    ! finding the index of the lowest allowed velocity ! it has to be greater than the lowest velocity scanned
    index_min=1
    velocity = origin 
    do while (velocity .lt. vel_min)
      index_min = index_min+1
      velocity = origin + (index_min-1)*increment
    end do

    ! finding the index of the highest allowed velocity !it has to be smaller than the highest velocity scanned
    index_max=n
    velocity = origin + (n-1)*increment
    do while (velocity .gt. vel_max)
      index_max = index_max-1
      velocity = velocity-increment
    end do

    ! calculating mean
    mean=0.
    total_weight=0.
    do i=index_min,index_max
      velocity = origin + (i-1)*increment
      mean = mean + array(i) * velocity
      total_weight = total_weight + array(i) 
    end do
    mean = mean/total_weight    

    ! calculating standard deviation
    sigma=0.
    do i=index_min,index_max
      velocity = origin + (i-1)*increment
      sigma = sigma + (velocity-mean)*(velocity-mean)*array(i)
    end do
    sigma = sqrt(sigma)/2.

    ! filling the values of the cdf for velocities between vel_min and vel_max
    do i=index_min,index_max
      velocity=origin+(i-1)*increment
      velocity=(velocity-mean)/sigma ! calculating the 'x'
      temp_sum=velocity
      new_term=velocity
      do k=1,200
        new_term=new_term*velocity*velocity/(2*k+1)
        temp_sum=temp_sum+new_term
      end do
      cdf(i)=0.5+temp_sum/sqrt(2*3.14)*exp(-velocity*velocity/2)
    end do 
    
    do i=index_max,n
      cdf(i)=1.
    end do

  end function



!  function raw_cdf(semblance,origin,increment,vel_min,vel_max) result(cdf) ! gets the cdf for a 1D array
!    real,dimension(:)                     :: semblance ! array can be the semblance values for instance
!    real,dimension(:),allocatable         :: cdf
!    integer                               :: i,n,total_nb_samples
!    integer,dimension(:),allocatable      :: occurrence ! array containing the nb of occurrence for each velocity value evaluated by the semblance plot
!    real                                  :: origin,increment,vel_min,vel_max,velocity,total_weight
!    integer                               :: index_min,index_max

!    n=size(semblance)
!    allocate(occurrence(n))

    ! finding the index of the lowest allowed velocity ! it has to be greater than the lowest velocity scanned
!    index_min=1
!    velocity = origin 
!    do while (velocity .lt. vel_min)
!      index_min = index_min+1
!      velocity = origin + (index_min-1)*increment
!    end do

    ! finding the index of the highest allowed velocity !it has to be smaller than the highest velocity scanned
!    index_max=n
!    velocity = origin + (n-1)*increment
!    do while (velocity .gt. vel_max)
!      index_max = index_max-1
!      velocity = velocity-increment
!    end do

    ! finding the total weighting (ie sum of all semblance values for a given tau)
!    total_weight=0.
!    do i=index_min,index_max
!      total_weight = total_weight + semblance(i)
!    end do

    ! normalizing the semblance values by the sum of all semblance values and attributing a occurence (integer) for each velocity
    !total_nb_samples=0
    !do i=index_min,index_max
    !  occurrence(i)=array(i)/total_weight*100000 ! occurence vector is a array of integer 
    !  total_nb_samples = total_nb_samples+occurrence(i)
    !end do

    ! creating the cdf
!    allocate(cdf(n))
!    cdf=0.
!    cdf(index_min)=semblance(index_min)
!    do i=index_min+1,index_max
!      cdf(i)=cdf(i-1) + semblance(i)
!    end do
!    cdf = cdf/total_weight
!    do i=index_max,n
!      cdf(i)=1.
!    end do 
    
!  end function


  
!  function generate_velocity_pick_gaussian(array,it,o1,o2,d1,d2,min_val,max_val) result(velocity_pick)
!    real                                :: velocity_pick,rand_nb,dval,d1,d2,o1,o2,min_val,max_val
!    real,dimension(:)                   :: array
!    integer                             :: it,counter,nval
!    real,dimension(:),allocatable       :: cdf
    
!    dval=d2
!    nval=(max_val-min_val)/dval
!    cdf = build_gaussian_cdf(array,o1,o2,d1,d2,min_val,max_val)
!    call random_number(rand_nb)
!    counter=1
!    value=cdf(it,counter)
!    if (value .ge. rand_nb) then
!      velocity_pick=min_val
!    else
!      do while ((value .lt. rand_nb) .and. (counter .le. nval))
!        value = cdf(it,counter) 
!        counter=counter+1
!      end do
!      if (counter .gt. nval) then
!        velocity_pick=max_val
!      else  
!      velocity_pick=min_val+(counter-1)*dval
!      end if
!    end if
!  end subroutine  

end module
